# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import math
from skoolparser import SkoolParser

# The ID of the main disassembly
MAIN_CODE_ID = 'main'

class SkoolGame:
    def __init__(self, skoolfile, ref_parser, code_id=MAIN_CODE_ID):
        self.ref_parser = ref_parser
        self.info = self.ref_parser.build_dictionary('Info')
        self.paths = self.ref_parser.build_dictionary('Paths')
        self.titles = self.ref_parser.build_dictionary('Titles')
        self.links = self._parse_links(self.ref_parser.build_dictionary('Links'))
        self.features = self.ref_parser.build_dictionary('Features')
        self.other_code = self.ref_parser.build_dictionaries('OtherCode')
        self.code_path = self.get_code_path(code_id)
        self.game_vars = self.ref_parser.build_dictionary('Game')
        self.colours = self._get_colours(ref_parser.build_dictionary('Colours'))
        self.min_lesson = int(self.game_vars.get('MinLesson', 0))
        self.max_lesson = int(self.game_vars.get('MaxLesson', 0))
        self.min_character = int(self.game_vars.get('MinCharacter', 0))
        self.max_character = int(self.game_vars.get('MaxCharacter', 0))
        self.char_buffer_base_byte = int(self.game_vars.get('CharBufferBaseByte', 0))
        self.game = self.game_vars.get('Game', skoolfile[:-6])
        self.play_area_size = [int(i) for i in self.game_vars.get('PlayAreaSize', '96,21').split(',')]
        self.play_area_top_left = [int(i) for i in self.game_vars.get('PlayAreaTopLeft', '0,0').split(',')]
        self.font_address = int(self.game_vars.get('FontAddress', 0))
        self.tap_info = self._get_tap_info()
        self.tap_address_table = self.tap_info.tap_address_table if self.tap_info else None

        self.skoolfile = skoolfile
        self.as_descs = self.ref_parser.build_dictionary('AnimatoryStates')
        self.characters = self.ref_parser.build_dictionary('Characters')
        self.lesson_descs = self.ref_parser.build_dictionary('Lessons')
        self.keypress_routines = self.ref_parser.build_dictionary('KeypressRoutines')
        self.map_details = self.ref_parser.build_dictionary('MapDetails')
        self.message_addresses = {}

        self.parser = SkoolParser(skoolfile, tap_info=self.tap_info)
        self.snapshot = self.parser.snapshot
        self._snapshots = [(self.snapshot, '')]
        self.entries = self.parser.entries
        self.memory_map = [e for e in self.parser.memory_map if e.ctl != 'i']
        self.tap_addresses = self.parser.tap_addresses
        self.tap_descs = self.parser.tap_descs

    def _get_colours(self, cdict):
        colours = {}
        for k, v in cdict.items():
            colours[k] = tuple([int(n) for n in v.split(',')])
        return colours

    def _get_tap_info(self):
        return None

    def _parse_links(self, links):
        new_links = {}
        for page_id, link_text in links.items():
            if link_text.startswith('['):
                sb_count = 1
                index = 1
                length = len(link_text)
                while index < length and sb_count > 0:
                    if link_text[index] == '[':
                        sb_count += 1
                    elif link_text[index] == ']':
                        sb_count -= 1
                    index += 1
                if sb_count == 0:
                    new_links[page_id] = (link_text[1:index - 1], link_text[index:])
            if page_id not in new_links:
                new_links[page_id] = (link_text.strip(), '')
        return new_links

    def get_code_path(self, code_id):
        for c_id, code in self.other_code:
            if c_id == code_id:
                return code['Path']
        return 'asm'

    def has_feature(self, feature):
        return self.features.get(feature, 0)

    def has_gbuffer(self):
        for entry in self.memory_map:
            if entry.ctl == 'g':
                return True
        return False

    def has_ref_items(self, section_type):
        return self.ref_parser.has_ref_items(section_type)

    def get_section(self, section_name):
        return self.ref_parser.get_section(section_name)

    def get_ref_items(self, section_type, keep_lines=False, as_list=False):
        return self.ref_parser.get_ref_items(section_type, keep_lines, as_list)

    def get_list(self, section_name, default=()):
        return self.ref_parser.get_list(section_name) or default

    def get_entry(self, address):
        return self.parser.get_entry(address)

    def get_entry_point_refs(self, address):
        return self.parser.get_entry_point_refs(address)

    def get_tap_refs(self, address):
        return self.parser.get_tap_refs(address)

    def get_container(self, address):
        return self.parser.get_container(address)

    def get_snapshot_name(self):
        return self._snapshots[-1][1]

    def pop_snapshot(self):
        self.snapshot = self._snapshots.pop()[0]

    def push_snapshot(self, name=''):
        self._snapshots.append((self.snapshot[:], name))

    def _get_screen_udg(self, row, col):
        attr = self.snapshot[22528 + 32 * row + col]
        address = 16384 + 2048 * (row // 8) + 32 * (row % 8) + col
        return Udg(attr, self.snapshot[address:address + 2048:256])

    def get_screenshot(self, x=0, y=0, w=32, h=24):
        width = min((w, 32 - x))
        height = min((h, 24 - y))
        scr_udgs = []
        for r in range(y, y + height):
            scr_udgs.append([self._get_screen_udg(r, c) for c in range(x, x + width)])
        return scr_udgs

    def _get_box_udg(self, address, row, col, attr=None):
        if attr is None:
            attr = self.snapshot[address + 8 * row + col]
        start = address + 24 + 64 * row + col
        return Udg(attr, self.snapshot[start:start + 64:8])

    def get_box_graphic(self, address, attr=None):
        udg_array = []
        for row in range(3):
            udg_array.append([self._get_box_udg(address, row, col, attr) for col in range(8)])
        return udg_array

    def _flip_byte(self, byte):
        f = 0
        for b in range(8):
            f = 2 * f + (byte & 1)
            byte /= 2
        return f

    def _flip_udg(self, udg):
        for index in range(8):
            udg.data[index] = self._flip_byte(udg.data[index])
            udg.mask[index] = self._flip_byte(udg.mask[index])

    def _get_sprite_udg(self, state, attr, ref_page, udg_page):
        ref_addr = state + 256 * ref_page
        ref = self.snapshot[ref_addr]
        if ref == 0:
            udg = [0] * 8
            mask = [255] * 8
        else:
            udg_addr = ref + 256 * udg_page
            udg = self.snapshot[udg_addr:udg_addr + 4096:512]
            mask = self.snapshot[udg_addr + 256:udg_addr + 4352:512]
        return Udg(attr, udg, mask, ref_addr=ref_addr, ref=ref, udg_page=udg_page)

    def _get_sprite_size(self, state):
        return 3, 4

    def build_sprite(self, state, attr=None, udg_page=None):
        width, height = self._get_sprite_size(state)
        udg_array = []
        for row in range(height):
            udg_array.append([self.get_sprite_udg(state, attr, row, col, udg_page) for col in range(width)])
        return udg_array

    def get_font_bitmap(self, character):
        base_address = ord(character) + self.font_address + 256
        width = self.snapshot[base_address - 256]
        return self.snapshot[base_address:base_address + 256 * width:256]

    def get_text(self, words):
        columns = []
        for character in words:
            columns.append(0)
            columns.extend(self.get_font_bitmap(character))
        udg_array = []
        for col in range(len(columns)):
            col_byte = columns[col]
            udg_index = col // 8
            bit = int(math.pow(2, 7 - (col - 8 * udg_index)))
            if udg_index == len(udg_array):
                udg_array.append([0] * 8)
            udg = udg_array[udg_index]
            for b in range(8):
                udg[7 - b] |= bit * (col_byte & 1)
                col_byte /= 2
        return udg_array

    def _draw_text(self, udgs, words, x, y):
        text_udgs = self.get_text(words)
        for col in range(len(text_udgs)):
            udg = udgs[y][col + x].data
            for index in range(8):
                udg[index] |= text_udgs[col][index]

    def get_skool_udgs(self, x, y, w, h, show_chars=False, show_x=0):
        skool_udgs = []
        for row in range(y, y + h):
            skool_udgs.append([])
            for col in range(x, x + w):
                skool_udgs[-1].append(self.get_skool_udg(row, col, show_chars))
        if show_x > 0:
            line = [Udg(8 * (7 - skool_x % 2), [0] * 8) for skool_x in range(x, x + w)]
            for skool_x in range(x, x + w):
                if skool_x % show_x == 0:
                    self._draw_text([line], str(skool_x), skool_x - x, 0)
            skool_udgs.insert(0, line)
            skool_udgs.append(line)
        return skool_udgs

    def _superimpose_sprite_udgs(self, skool_udg, x, y):
        for state, char_x, char_y in self.get_initial_states():
            if char_y <= y <= char_y + 3 and char_x <= x <= char_x + 2:
                sprite_udg = self.get_sprite_udg(state, None, y - char_y, x - char_x)
                for i in range(8):
                    skool_udg.data[i] = (skool_udg.data[i] | sprite_udg.data[i]) & sprite_udg.mask[i]

class SkoolDaze(SkoolGame):
    def _get_tap_info(self):
        tap_info = TapInfo(65152, 222)

        tap_info.set_num_params(25303, 1) # Move about until...
        tap_info.set_num_params(25534, 2) # Go to...
        tap_info.set_num_params(27453, 2) # Make character do...
        tap_info.set_num_params(27476, 1) # Unsignal that...
        tap_info.set_num_params(27480, 1) # Signal that...
        tap_info.set_num_params(31834, 2) # Put next address in character's buffer, making him...
        tap_info.set_num_params(31838, 2) # Put next address in character's buffer, making him...
        tap_info.set_num_params(32167, 1) # Write on board unless...

        tap_info.add_controller(31834)
        tap_info.add_controller(31838)

        tap_info.add_terminal_command(25247)
        tap_info.add_terminal_command(25303, [0])
        tap_info.add_terminal_command(31162)
        tap_info.add_terminal_command(31815)
        tap_info.add_terminal_command(31854)
        tap_info.add_terminal_command(32178)
        tap_info.add_terminal_command(63964)
        tap_info.add_terminal_command(64035)
        return tap_info

    def get_skool_udg(self, y, x, show_chars=False):
        ref_page = y + 152
        attr_addr = x + 128 + 256 * ref_page
        ref_addr = x + 256 * ref_page
        ref = self.snapshot[ref_addr]
        attr = self.snapshot[attr_addr]
        udg_page = 128 + 8 * (x // 32)
        udg_addr = ref + 256 * udg_page
        udg_bytes = self.snapshot[udg_addr:udg_addr + 2048:256]
        skool_udg = Udg(attr, udg_bytes, attr_addr=attr_addr, ref_addr=ref_addr, ref=ref, udg_page=udg_page, x=x, y=y)
        if show_chars:
            self._superimpose_sprite_udgs(skool_udg, x, y)
        return skool_udg

    def get_sprite_udg(self, state, attr, row, col, udg_page=None):
        if udg_page is None:
            udg_page = 201 if 71 < (state & 127) < 104 else 185
        ref_page = 173 + col * 4 + row
        return self._get_sprite_udg(state, attr, ref_page, udg_page)

    def get_initial_states(self):
        initial_states = []
        for c in range(18) + [20]:
            initial_states.append((self.snapshot[56064 + c], self.snapshot[56320 + c], 17))
        return initial_states

class BackToSkool(SkoolGame):
    def __init__(self, skoolfile, ref_parser, code_id=MAIN_CODE_ID):
        SkoolGame.__init__(self, skoolfile, ref_parser, code_id)
        for n in range(112) + range(128, 208):
            self.message_addresses[str(n)] = str(self.snapshot[n + 65024] + 256 * self.snapshot[n + 65280])

    def _get_tap_info(self):
        tap_info = TapInfo(59392, 88)

        tap_info.set_num_params(25700, 2) # Go to...
        tap_info.set_num_params(29052, 2) # Open/close door/window
        tap_info.set_num_params(31952, 1) # Move about until...
        tap_info.set_num_params(31969, 2) # Move about until...
        tap_info.set_num_params(32497, 1) # Write on board unless...
        tap_info.set_num_params(61555, 1) # Restart command list or tell kids to sit down and signal that...
        tap_info.set_num_params(62599, 1) # Move forward 10 places in command list unless...
        tap_info.set_num_params(63447, 2) # Put next address in character's buffer, making him...

        tap_info.add_controller(63447)

        tap_info.add_terminal_command(31952, [0])
        tap_info.add_terminal_command(53796)
        tap_info.add_terminal_command(61440)
        tap_info.add_terminal_command(61624)
        tap_info.add_terminal_command(61696)

        return tap_info

    def get_skool_udg(self, y, x, show_chars=False):
        q_addr = x + 46336
        q = self.snapshot[q_addr]
        h = y + 160
        hl = 180 + q // 2 + 256 * h
        i = 0
        if q & 1:
            p = self.snapshot[hl] & 15
        else:
            p = (self.snapshot[hl] & 240) // 16
        if p == 0:
            i = q & 3 or 2
            if q & 1:
                p = self.snapshot[hl + 256] & 15
            else:
                p = (self.snapshot[hl + 256] & 240) // 16
        attr = i | (8 * p)
        ref_addr = q + 256 * h
        ref = self.snapshot[ref_addr]
        a = 136 // int(math.pow(2, q & 3)) & self.snapshot[144 + q // 4 + 256 * h]
        udg_page = (8 if a & 15 else 0) + (144 if a & 240 else 128)
        udg_addr = ref + 256 * udg_page
        udg_bytes = self.snapshot[udg_addr:udg_addr + 2048:256]
        skool_udg = Udg(attr, udg_bytes, q_addr=q_addr, ref_addr=ref_addr, ref=ref, udg_page=udg_page, x=x, y=y)
        if show_chars:
            self._superimpose_sprite_udgs(skool_udg, x, y)
        return skool_udg

    def get_sprite_udg(self, state, attr, row, col, udg_page=None):
        as_norm = state | 128
        udg_page = 183 if as_norm < 208 else 199
        col = 2 - col if state & 128 else col
        ref_page = 215 + col * 4 + row
        udg = self._get_sprite_udg(as_norm, attr, ref_page, udg_page)
        if state & 128:
            self._flip_udg(udg)
        return udg

    def _alter_skool_udgs(self, address):
        coords = []
        while True:
            y = self.snapshot[address]
            if y == 255:
                break
            x = self.snapshot[address + 1]
            coords.append((x, y))
            udg_lsb = self.snapshot[address + 2]
            attr = self.snapshot[address + 3]
            d = y + 160
            col_ref = self.snapshot[x + 46336]
            self.snapshot[col_ref + 256 * d] = udg_lsb
            z = int(math.pow(2, col_ref & 3))
            e = 128 + (col_ref + 104) // 2
            a = self.snapshot[e + 256 * d]
            if col_ref & 1:
                a = (a & 240) | (attr & 15)
            else:
                a = (16 * (attr & 15)) | (a & 15)
            self.snapshot[e + 256 * d] = a
            a = attr & 192
            if a & 64:
                a = a - 56
            e = 144 + col_ref // 4
            self.snapshot[e + 256 * d] = self.snapshot[e + 256 * d] & 255 - 136 // z | a // z
            address += 4
        return coords

    def _get_skool_udgs_by_coords(self, coords):
        left_x, top_y, width, height = self._get_dimensions(coords)
        udg_array = []
        for y in range(top_y, top_y + height):
            udg_array.append([None] * width)
        for x, y in coords:
            udg_array[y - top_y][x - left_x] = self.get_skool_udg(y, x)
        return udg_array

    def _get_dimensions(self, coords, padding=0):
        x_range = [c[0] for c in coords]
        y_range = [c[1] for c in coords]
        left_x = min(x_range)
        top_y = min(y_range)
        width = max(x_range) - left_x + 1
        height = max(y_range) - top_y + 1
        left_x -= padding
        top_y -= padding
        width += 2 * padding if left_x + width < 192 else padding
        height += 2 * padding if top_y + height < 20 else padding
        return left_x, top_y, width, height

    def get_udg(self, address, attr):
        return Udg(attr, self.snapshot[address:address + 8])

    def get_mutable_udg_array(self, address, padding):
        self.push_snapshot()
        coords = self._alter_skool_udgs(address)
        if padding >= 0:
            udg_array = self.get_skool_udgs(*self._get_dimensions(coords, padding))
        else:
            udg_array = self._get_skool_udgs_by_coords(coords)
        self.pop_snapshot()
        return udg_array

    def get_initial_states(self):
        return [self.snapshot[32 + 256 * c:35 + 256 * c] for c in range(183, 211)]

class ContactSamCruise(SkoolGame):
    def __init__(self, skoolfile, ref_parser, code_id=MAIN_CODE_ID):
        SkoolGame.__init__(self, skoolfile, ref_parser, code_id)
        self.font = {
            48: (124, 138, 146, 124),
            49: (66, 254, 2),
            50: (70, 138, 146, 98),
            51: (130, 146, 178, 204),
            52: (56, 72, 254, 8),
            53: (242, 146, 146, 140),
            54: (124, 146, 146, 140),
            55: (128, 134, 152, 224),
            56: (108, 146, 146, 108),
            57: (98, 146, 146, 124)
        }

    def get_skool_udg(self, y, x, show_chars=False):
        xh = x // 8
        yhp = 256 * (184 + y // 6)
        t = self.snapshot[xh + yhp]
        m = self.snapshot[21 + 256 * (160 + (x % 8))]
        ylp = 256 * (160 + 2 * (y % 6))
        if self.snapshot[t + 48896] & m:
            ylp += 3072
        r = self.snapshot[t + 128 + 256 * (184 + (x % 8))]
        z = self.snapshot[r + ylp]
        attr_ref = self.snapshot[32 + xh + yhp]
        attr = self.snapshot[attr_ref + 256 * (168 + (z & 3))]
        ref = self.snapshot[r + ylp + 256]
        if z & 128:
            # This location has a background and a foreground
            if z & 64:
                l = 128 + (z & 60) // 2
                if ref & 128:
                    l += 1
                a = ref | 128
            else:
                j = self.snapshot[64 + xh + yhp]
                k = self.snapshot[192 + xh + 65280]
                if self.snapshot[k + 32512]:
                    j |= 32
                if z & 4:
                    j = 64 * (j & 3) + (j & 32)
                a = ref | (j & 224)
                l = 159
            udg_addr = l + 32768
            data = self.snapshot[udg_addr:udg_addr + 2048:256]
            if show_chars:
                self._superimpose_sprite_udgs(data, x, y, 1)
            l = self.snapshot[a + 49152]
            h = 128
            for i in range(4):
                data[i] |= self.snapshot[l + 256 * h]
                data[i] &= self.snapshot[l + 256 * (h + 1)]
                h += 2
            l = self.snapshot[a + 49408]
            h = 128
            for i in range(4, 8):
                data[i] |= self.snapshot[l + 256 * h]
                data[i] &= self.snapshot[l + 256 * (h + 1)]
                h += 2
        else:
            udg_addr = ref + 256 * (128 + 2 * (z & 28))
            data = self.snapshot[udg_addr:udg_addr + 2048:256]
            if show_chars and z & 64 == 0:
                self._superimpose_sprite_udgs(data, x, y, 1)

        if show_chars:
            self._superimpose_sprite_udgs(data, x, y, 2)
            self._superimpose_sprite_udgs(data, x, y, 4)
        return Udg(attr, data)

    def _superimpose_sprite_udgs(self, skool_udg_data, x, y, z):
        initial_states = [self.snapshot[a:a + 4] for a in range(65198, 65408, 6)]
        initial_states += [self.snapshot[a:a + 4] for a in range(56870, 58918, 256)]
        sam_z = self.snapshot[58884]
        if sam_z == 0:
            sam_as, sam_x, sam_y, sam_z = 0, 229, 19, 1
        else:
            sam_as, sam_x, sam_y = self.snapshot[58880:58883]
        initial_states.append((sam_as, sam_x, sam_y, sam_z))
        for state, char_x, char_y, char_z in initial_states:
            if char_z == z:
                width, height = self._get_sprite_size(state)
                if char_y <= y < char_y + height and char_x <= x < char_x + width:
                    sprite_udg = self.get_sprite_udg(state, None, y - char_y, x - char_x)
                    for i in range(8):
                        skool_udg_data[i] = (skool_udg_data[i] | sprite_udg.data[i]) & sprite_udg.mask[i]

    def _get_sprite_size(self, state):
        width = 3
        height = 5
        if state & 127 >= 120:
            width = height = 2
        elif state % 8 == 7:
            width = 5
            height = 3
        return width, height

    def get_sprite_udg(self, state, attr, row, col, udg_page=None):
        width, height = self._get_sprite_size(state)
        as_norm = (state if state & 7 else state + 2) | 128
        udg_page = 199
        col = width - 1 - col if state & 128 else col
        ref_page = 215 + col * height + row
        udg = self._get_sprite_udg(as_norm, attr, ref_page, udg_page)
        if state & 128:
            self._flip_udg(udg)
        return udg

    def get_font_bitmap(self, character):
        return self.font.get(ord(character), (255, 255, 255, 255))

    def get_initial_states(self):
        return ()

class Udg:
    def __init__(self, attr, data, mask=None, attr_addr=None, q_addr=None, ref_addr=None, ref=None, udg_page=None, x=None, y=None):
        self.attr = attr
        self.data = data
        self.mask = mask
        self.attr_addr = attr_addr
        self.q_addr = q_addr
        self.ref_addr = ref_addr
        # We store the UDG reference now in case the snapshot changes before the
        # reference is looked up via self.snapshot[udg.ref_addr] (e.g. because a
        # saved snapshot is restored after alter_skool_udgs() has been called)
        self.ref = ref
        self.udg_page = udg_page
        self.x = x
        self.y = y
        # If ink or paper is set, it will override attr (see
        # skoolhtml.SkoolWriter.get_pixel_array())
        self.ink = None
        self.paper = None

def get_skool_class(name):
    if name == 'SkoolDaze':
        return SkoolDaze
    if name == 'BackToSkool':
        return BackToSkool
    if name == 'ContactSamCruise':
        return ContactSamCruise
    return SkoolGame

class TapInfo:
    def __init__(self, tap_address_table, max_tap):
        self.tap_address_table = tap_address_table
        self.min_tap = tap_address_table % 256
        self.max_tap = max_tap
        self.tap_routine_params = {}
        self.tap_controllers = []
        self.terminal_commands = {}

    def get_tap_range(self):
        return range(self.min_tap, self.max_tap + 1, 2)

    def get_num_params(self, address):
        return self.tap_routine_params.get(address) or 0

    def is_terminal_command(self, address, params):
        return address in self.terminal_commands and self.terminal_commands[address] == params

    def is_controller(self, address):
        return address in self.tap_controllers

    def set_num_params(self, address, num_params):
        self.tap_routine_params[address] = num_params

    def add_controller(self, address):
        self.tap_controllers.append(address)

    def add_terminal_command(self, address, params=None):
        self.terminal_commands[address] = params or []
